from __future__ import annotations
from contextlib import suppress
from typing import Any, List, Iterable, Union


class FlaggedDict(dict):
    def __init__(self, *args, __flags__: Union[List, Iterable] = None, **kwargs):  # noqa
        super().__init__(*args, **kwargs)
        self.flags = {key: None for key in __flags__} if __flags__ else {}

    def clear(self) -> None:
        super().clear()
        self.flags.clear()

    def copy(self) -> FlaggedDict:
        return FlaggedDict(self, __flags__=self.flags)

    def pop(self, k, *args) -> Any:
        with suppress(KeyError):
            self.flags.pop(k)

        return super().pop(k, *args)

    def popitem(self) -> tuple:
        k, v = super().popitem()
        with suppress(KeyError):
            self.flags.pop(k)
        return k, v

    def setdefault(self, k, d=None) -> Any:
        if k not in self:
            self.flags[k] = None

        return super().setdefault(k, d)

    def update(self, *args, **kwargs) -> None:
        super().update(*args, **kwargs)
        for k in kwargs:
            self.flags[k] = None

    def __delitem__(self, k) -> None:
        with suppress(KeyError):
            self.flags.pop(k)
        super().__delitem__(k)

    def __repr__(self):
        return f"FlaggedDict({super().__repr__()}, __flags__={list(self.flags.keys())})"

    def __setitem__(self, k, v) -> None:
        self.flags[k] = None
        super().__setitem__(k, v)



import ast
import uuid
from pathlib import Path
from types import SimpleNamespace


def gen_sym(varname):
    uuid_str = str(uuid.uuid4()).replace("-", "")
    return f"__{varname}_{uuid_str}__"


def add_asignment_to_last_statement(code: ast.Module, varname: str):
    # If the last statement is not an Expression, don't do anything
    if code.body and isinstance(code.body[-1], ast.Expr):

        assign = ast.parse(f"{varname}=None").body[0]
        last_expression = getattr(code.body[-1], "value", code.body[-1])

        assign.value = last_expression
        for attr in ("lineno", "col_offset", "end_lineno", "end_col_offset"):
            setattr(assign, attr, getattr(code.body[-1], attr))

        code.body[-1] = assign


def format_code_for_error_line_display(code: str, lineno: int, filename: str):
    lineno = int(lineno)
    strlineno = str(lineno)

    lines = code.splitlines()
    lines_annotated = [f"{i+1:7d} {line}" for i, line in enumerate(lines)]
    lines_annotated[lineno - 1] = ("-") * (5 - len(strlineno)) + "> " + strlineno + " " + lines[lineno - 1]
    return "\n".join([f"Error in {Path(filename).name}:"] + lines_annotated)


def deepest_traceback(traceback, filename):
    pointer = SimpleNamespace(tb_next=traceback)
    tb_find = None
    while (getattr(pointer, "tb_next", None)) is not None:
        pointer = pointer.tb_next
        if pointer.tb_frame.f_code.co_filename == filename:
            tb_find = pointer

    return tb_find



import unittest
from pathlib import Path
from textwrap import dedent
import sys

this_dir = Path(__file__).resolve().parent
sys.path.insert(0, this_dir.parent.as_posix())

from neval import flagged_dict
from neval import neval

FlaggedDict = flagged_dict.FlaggedDict


class TestFlaggedDict(unittest.TestCase):
    def setUp(self) -> None:
        self.d = FlaggedDict(
            {"d": 4, "e": 5, "f": 6, "a": 1, "b": 2, "c": 3},
            __flags__={"a": None, "b": None, "c": None},
        )

    def test_each_constructors(self):
        self.assertEqual(dict(**self.d), {"d": 4, "e": 5, "f": 6, "a": 1, "b": 2, "c": 3})

        self.assertEqual(self.d.get("a"), 1)

        self.assertEqual(dict.fromkeys(["a", "b"], 2), FlaggedDict.fromkeys(["a", "b"], 2))

    def test_pops(self):
        self.assertEqual(self.d.pop("c"), 3)

    def test_popitem(self):
        self.assertEqual(self.d.popitem(), ("c", 3))
        self.assertEqual(self.d.popitem(), ("b", 2))
        self.assertEqual(self.d.popitem(), ("a", 1))
        self.assertEqual(self.d.popitem(), ("f", 6))

    def test_clear(self):
        self.d.clear()
        self.assertEqual(self.d, {})

    def test_copy(self):
        d = self.d.copy()
        self.assertEqual(d, self.d)

        d.pop("c")
        self.assertNotEqual(d, self.d)

    def test_items_keys_values(self):
        # items
        self.assertEqual(
            list(self.d.items()),
            [("d", 4), ("e", 5), ("f", 6), ("a", 1), ("b", 2), ("c", 3)],
        )

        self.assertEqual(len(self.d.items()), 6)

        self.assertEqual(list({**self.d}.items()), list(self.d.items()))

        self.assertEqual({**self.d}.items(), self.d.items())

        # test contains
        self.assertTrue(("a", 1) in dict(**self.d).items())
        self.assertTrue(("a", 1) in self.d.items())

        # keys
        self.assertEqual(list(self.d.keys()), ["d", "e", "f", "a", "b", "c"])

        self.assertEqual(len(self.d), 6)

        self.assertEqual(list({**self.d}.keys()), list(self.d.keys()))

        self.assertEqual({**self.d}.keys(), self.d.keys())

        self.assertIn("a", self.d.keys())

        # values
        self.assertEqual(list(self.d.values()), [4, 5, 6, 1, 2, 3])

        self.assertEqual(len(self.d.values()), 6)

        self.assertEqual(list({**self.d}.values()), list(self.d.values()))

        # self.assertEqual({**self.d}.values(), self.d.values())

        self.assertTrue(6 in self.d.values())

    def test_set_default(self):
        self.assertEqual(self.d.setdefault("a", 999), 1)
        self.assertEqual(self.d.setdefault("g", 999), 999)

    def test_update(self):
        self.d.update({"a": 999, "g": 999})
        self.assertEqual(self.d, {"d": 4, "e": 5, "f": 6, "a": 999, "b": 2, "c": 3, "g": 999})

    def test_delete(self):
        del self.d["a"]
        self.assertEqual(self.d, {"d": 4, "e": 5, "f": 6, "b": 2, "c": 3})

    def test_iter(self):
        self.assertEqual(list(self.d), ["d", "e", "f", "a", "b", "c"])

    def test_len(self):
        self.assertEqual(len(self.d), 6)

    def test_reverse(self):
        self.assertEqual(list(reversed(self.d)), ["c", "b", "a", "f", "e", "d"])


class TestFlaggedDictExec(unittest.TestCase):
    def test_annotation(self):
        code = dedent(
            """
            a = object()
            class _:
                def _(self, _=a): pass
            """
        )
        exec(code, {})
        exec(code, FlaggedDict())


class TestRunCode(unittest.TestCase):
    def test_assignment(self):

        self.assertEqual(2, neval("1+1"))

        self.assertEqual(None, neval("for i in range(10): i"))

        self.assertEqual(
            9,
            neval(
                dedent(
                    """
                    for i in range(10):
                        i
                    i
                    """
                )
            ),
        )

        self.assertEqual(
            None,
            neval("while True: break"),
        )

    def test_return(self):
        self.assertEqual(
            # Do we want to return None or the function `f`?
            None,
            neval("def f(): pass"),
        )

        self.assertTrue(callable(neval("lambda:None")))

        # unfortunately the `with` statement is the last statement and not the `1` expression
        self.assertEqual(
            None,
            neval(
                dedent(
                    """
                    class w:
                        def __enter__(self): pass
                        def __exit__(self, *args): pass
                            
                    with w():
                        1
                    """
                )
            ),
        )

        self.assertEqual(None, neval(""))

    def test_scoping(self):

        # Exec used class definition scope when globals and locals are specified
        self.assertRaisesRegex(
            NameError,
            "name 'a' is not defined",
            lambda: exec("a=[1];[i for i in a if a]", {}, {}),
        )

        # Runcode does not use class definition scope when namespace (~locals) and
        # namespace_readonly (~globals) are specified
        self.assertEqual([1], neval("a=[1];[i for i in a if a]"))

        self.assertEqual(
            (None, {"a": 1, "b": 2}),
            (
                neval("a=1;b=2;", namespace := {}, {"a": 0}),
                namespace,
            ),
        )

        self.assertEqual(
            (None, {"a": 0}),
            (
                neval("a=1;b=2;", {}, namespace_readonly := {"a": 0}),
                namespace_readonly,
            ),
        )

        self.assertEqual(
            (None, ["mock_module_54f56", "os", "sys"]),
            (
                neval(
                    dedent(
                        f"""
                        import sys
                        import os
                        sys.path.insert(
                            0,
                            os.path.join({repr(str(Path(__file__).resolve().parent))}, "import_examples")
                        )
                        import mock_module_54f56;
                        """
                    ),
                    namespace := {},
                    {},
                ),
                sorted(namespace),
            ),
        )

        import os
        import sys

        self.assertEqual(
            (None, ["mock_module_54f56"]),
            (
                neval(
                    dedent(
                        f"""
                        sys.path.insert(
                            0,
                            os.path.join({repr(str(Path(__file__).resolve().parent))}, "import_examples")
                        )
                        import mock_module_54f56;
                        """
                    ),
                    namespace := {},
                    {"os": os, "sys": sys},
                ),
                sorted(namespace),
            ),
        )

        self.assertEqual(
            (None, ["Example", "example"]),
            (
                neval(
                    dedent(
                        f"""
                        Example = object()
                        def example(a=Example):
                            pass
                        """
                    ),
                    namespace := {},
                    {},
                ),
                sorted(namespace),
            ),
        )

        self.assertEqual(
            (None, sorted([i for i in dir(flagged_dict) if not i.startswith("_")])),
            (
                neval(
                    this_dir.parent.joinpath("neval", "flagged_dict.py").read_text(),
                    namespace := {},
                ),
                sorted(namespace),
            ),
        )

        # Do we want this to be true or not? For now it's False
        self.assertEqual(
            False,
            neval(
                dedent(
                    """
                        name_main = False
                        if __name__ == "__main__":
                            name_main = True
                        name_main
                        """
                ),
                namespace := {},
            ),
        )

        # Assign-and-return via walrus operator
        self.assertEqual(
            (1, {"a": 1}),
            (neval("(a := 1)", namespace := {}), namespace),
        )

    def test_errors(self):
        code = dedent(
            """\
            a
            b
            c
            d d d
            e"""
        )
        #   File "<neval-2b43d9fb0bb36723ba251724e7078a6a33f6fefd>", line 1
        #     a = 1 b = 2
        #           ^
        # SyntaxError: invalid syntax
        # ----> 1 a = 1 b = 2
        #

        self.assertRaisesRegex(
            SyntaxError,
            "invalid syntax",
            lambda: neval(code),
        )

        # test if python running this test is > 3.11
        if sys.version_info >= (3, 11):
            #
            err = None
            try:
                neval(code)
            except SyntaxError as e:
                err = e

            self.assertEqual(
                err.msg,
                "invalid syntax",
            )
            self.assertEqual(
                err.__notes__,
                [
                    dedent(
                        """\
                              1 a
                              2 b
                              3 c
                        ----> 4 d d d
                              5 e"""
                    )
                ],
            )

        self.assertRaisesRegex(
            ZeroDivisionError,
            "division by zero",
            lambda: neval("1/0"),
        )

        self.assertRaisesRegex(
            ZeroDivisionError,
            "division by zero",
            lambda: neval("1/0", {}, {}, False),
        )


if __name__ == "__main__":
    unittest.main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="neval",
    author="AutoActuary",
    description="Execute raw Python code in an isolated namespace",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AutoActuary/neval",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Other/Proprietary License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.4",
    use_scm_version={
        "write_to": "neval/version.py",
    },
    setup_requires=[
        "setuptools_scm",
    ],
    install_requires=[],
)



import ast
import hashlib
import linecache
import tempfile
from collections.abc import Mapping
from contextlib import suppress
from pathlib import Path
from types import SimpleNamespace
from typing import Union, Optional, Any
import re
from .flagged_dict import FlaggedDict
from .util import (
    gen_sym,
    add_asignment_to_last_statement,
    deepest_traceback,
    format_code_for_error_line_display,
)

reg_neval_filename = re.compile(r"neval-[0-9a-f]{40}")
neval_filename_cache = {}


def neval(
    code: Union[str, ast.Module],
    namespace: Optional[Union[Mapping, SimpleNamespace, Any]] = None,
    namespace_readonly: Optional[Union[Mapping, SimpleNamespace, Any]] = None,
    traceback_file_output: bool = True,
) -> Any:
    """
    Execute Python code in a namespace and return the result of the last statement in the code.

    Args:
        code (Union[str, ast.Module]): The code to execute.
        namespace (Union[Mapping, Any], optional): The namespace to execute the code in, this can either be a `dict` or
            any object with a `__dict__` attribute such as `SimpleNamespace`. Defaults to `None`.
        namespace_readonly (Union[Mapping, Any], optional): A namespace to execute the code in, this can
            either be a `dict` or any object with a `__dict__` attribute such as `SimpleNamespace`. Defaults to `None`.
        traceback_file_output (bool, optional): Option to dump the `code` string to a temporary file when an error
            occurs in order for the Python stacktrace to print all relevant lines. Ideally this should be mocked in
            memory, but it seems like there are some redundancies in the interpreter that doesn't make this process
            easy. This might cause security issues. Defaults to `True`.


    Returns:
        Any: The result of the last statement in the code. If that statement is not an expression None is returned.

    """

    def get_namespace_mapping(x):
        return {} if x is None else x if isinstance(x, Mapping) else x.__dict__

    nspace = get_namespace_mapping(namespace)

    # __builtins__ gets automatically injected by exec, remove it if it's not already there
    remove_builtins_from_namespace = "__builtins__" not in nspace

    # Combine the namespaces into one and flag the read/write ones
    ns_exec = FlaggedDict(nspace, __flags__=nspace)
    ns_exec.update(get_namespace_mapping(namespace_readonly))

    # Return the last statement to this unique variable
    var_return = gen_sym("return")

    # Create a fake traceback file to display the correct number of the error
    filename = Path(
        tempfile.gettempdir() if traceback_file_output else "",
        f"neval-{hashlib.sha1(str(code).encode('utf-8')).hexdigest()}",
    ).as_posix()
    neval_filename_cache[Path(filename).name] = filename

    # Set up the AST node
    runme = code

    # If a syntax error occurs, rather raise it at the exec line
    with suppress(SyntaxError):
        if not isinstance(runme, ast.AST):
            runme = ast.parse(runme)

        add_asignment_to_last_statement(runme, var_return)

    # Execute the annotated AST node
    try:
        exec(compile(runme, filename, "exec"), ns_exec)

    # If an error occurs, display it properly
    except Exception as e:
        lineno = None

        # `SyntaxError` is special since it is thrown by `compile` and not `exec`
        if isinstance(e, SyntaxError):
            e.filename = filename
            lineno = e.lineno

        # Inject fake file contents to traceback cache
        if isinstance(code, str):
            # Python 3.11 has new functionality to display traceback notes
            if hasattr(e, "add_note"):
                if lineno is None:
                    if tb_last := deepest_traceback(e.__traceback__, filename):
                        lineno = tb_last.tb_lineno
                if lineno is not None:
                    e.add_note(
                        format_code_for_error_line_display(code, lineno, filename),
                    )

        # Add content to temp in order for C to print the correct line number after Python teardown
        if traceback_file_output and isinstance(code, str):
            for fname in Path(tempfile.gettempdir()).glob("neval-*"):
                if reg_neval_filename.match(fname.name) and fname.name not in neval_filename_cache:
                    fname.unlink()

            Path(filename).write_text(code)

        raise

    # Even if an error occurs, ensure that mutated scope is reflected in the namespace
    finally:
        return_value = ns_exec.pop(var_return, None)
        if remove_builtins_from_namespace:
            ns_exec.pop("__builtins__", None)

        nspace.clear()

        for key in ns_exec.flags:
            nspace[key] = ns_exec[key]

    linecache.cache.pop(filename, None)

    return return_value
